package de.hawhamburg.reachability.sensor.adapter;

import org.apache.log4j.Logger;

import com.google.gson.Gson;

import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.sensor.Sensor;
import de.hawhamburg.reachability.sensor.SensorAdapter;
import de.hawhamburg.reachability.sensor.SensorAdapterFactory;
import de.hawhamburg.reachability.sensor.SensorAgent;
import de.hawhamburg.reachability.sensor.SensorData;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;
import de.hawhamburg.stresscompanion.output.JsonStressCompanionInformation;

/**
 * Sensor to determine the inhabitants level of stress.<br><br>
 * 
 * This sensor adapter adapts the stress level sensor, called StressCompanion (designed by Benjamin Lindemann). It 
 * receives the inhabitants current level of stress, which was previously calculated by the StressCompanion via an 
 * ActiveMQ result topic. The StressCompanion pushes frequently its calculations onto this topic. Due to this fact, no 
 * requests will be necessary.  
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class StressCompanionAdapter extends    Sensor
                                    implements SensorAdapter  {
        
    /**
     * Creates a new stress companion adapter which calculates the current stress level of the 
     * inhabitant while working on his computer.<br><br>
     * 
     * The sensors configurations will be loaded by the constructor of {@link Sensor} from the XML file in  
     * {@link de.hawhamburg.reachability.sensor.Settings#SENSOR_CONFIGURATION_PATH}
     * which is named like this class.
     * 
     * @param dimension The {@link ReasonerDimension} for which this sensor is configured.
     */
    public StressCompanionAdapter(final ReasonerDimension dimension) {
        super(dimension);
    }

    /**
     * Retrieves the current inhabitants stress level via ActiveMQ and returns it as {@link SensorData}.
     * 
     * @return          The current stress level as {@link SensorData} object. 
     */
    public final SensorData getSensorOutput() {
        String json = ConnectionUtil.getMessageOnTopic(this.getResultTopic());
        if (json != null) {
            double value = 0.0;
            
            Gson g = new Gson();
            JsonStressCompanionInformation msgObj = g.fromJson(json, JsonStressCompanionInformation.class);
            
            value = (double) msgObj.state.ordinal();
            OutputUtil.debug("Received stress level: " + msgObj.state + " (Weighted as " + value + ")");
            
            return new SensorData(this.getClass().getSimpleName(), this.getDataDimension(), value);
        } else {
            // No data found...
            return null;
        }
    }

    /**
     * Main method which starts this {@link SensorAdapter} wrapped by a {@link SensorAgent}.
     * 
     * @param args      The start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        Logger.getRootLogger().removeAllAppenders();
        
        // Build this sensor adapter by using the SensorAdapterFactory.
        SensorAdapter sa = SensorAdapterFactory.getConnectedSensorAdapter(StressCompanionAdapter.class.getSimpleName(),
                                                                          ReasonerDimension.stress);
        
        // Start the sensor adapter as sensor agent, which supplied the required control functions.
        new SensorAgent(sa);
    }
}
