package de.hawhamburg.reachability;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import de.hawhamburg.lifecycle.data.ConfigOrder;
import de.hawhamburg.lifecycle.data.JsonConfigOrder;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.properties.PropertyService;
import de.hawhamburg.reachability.properties.PropertyServiceNotifiable;
import de.hawhamburg.reachability.reasoner.enumeration.ReasonerDimension;
import de.hawhamburg.reachability.registry.Registry;
import de.hawhamburg.reachability.registry.enumeration.RegistryAgentType;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.EnumerationUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * The ReachabilityWorker is a thread which mainly serves as control service. It starts a property and a configuration
 * service and coordinates different distributed components of this systems. These components may be sensor or reasoner
 * adapters, service and user interfaces.
 *
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR}
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ReachabilityWorker extends     Thread
                                implements  StopableObject,
                                            PropertyServiceNotifiable {

    /**
     * Holds the activity state of this object. If active is <code>false</code>, the worker won't publish its results on
     *  the result topic.
     */
    private boolean active                                          = false;

    /**
     * Determines if the method {@link #shutDown()} has been called.
     */
    private boolean shutDown                                        = false;

    /**
     * Holds the information, if {@link #run()} statement has ended or not.
     */
    private boolean ended                                           = false;
    
    /**
     * The {@link LPPublisher} which sends control messages to other components via the ActiveMQ.
     */
    private LPPublisher globalControlTopic                          = null;

    /**
     * A {@link List} of registered reasoners.
     */
    private List<String> registeredReasoners                        = null;
    
    /**
     * The actual running reasoner.
     */
    private String runningReasoner                                  = null;

    /**
     * A {@link Map} of context dimensions ({@link ReasonerDimension}) to a {@link List} of sensors. This map holds
     * all registered sensor agents and their specific context dimensions.
     */
    private Map<ReasonerDimension, List<String>> registeredSensors  = null;
    
    /**
     * A {@link Map} of context dimensions ({@link ReasonerDimension}) to sensor agents, which contains all actual
     * running sensor agents and their specific context dimension.
     */
    private Map<ReasonerDimension, String> runningSensors           = null;

    /**
     * The property service thread {@link de.hawhamburg.reachability.properties.PropertyService}.
     */
    private PropertyService propertyService                         = null;

    /**
     * The registry service thread {@link de.hawhamburg.reachability.registry.Registry}.
     */
    private Registry registryService                                = null;

    /**
     * Constructs a new ReachabilityWorker object. It initializes different instance variables.
     */
    public ReachabilityWorker() {
        this.registeredReasoners    = new ArrayList<String>();
        this.registeredSensors      = new HashMap<ReasonerDimension, List<String>>();
        for (ReasonerDimension dim : ReasonerDimension.values()) {
            this.registeredSensors.put(dim, new ArrayList<String>());
        }
        this.runningSensors         = new HashMap<ReasonerDimension, String>();
    }

    /**
     * Method to set the active status {@link #active}. If the method is called with its parameter set to 
     * <code>false</code>, the worker will stop publishing its results on the result topic.
     *
     * @param pred                  If <code>false</code> the worker will stop publishing  results on the result topic.
     */
    public final void activ(final boolean pred) {
        this.active = pred;

        Map<ReasonerDimension, String> copy = new HashMap<ReasonerDimension, String>();
        for (Map.Entry<ReasonerDimension, String> sensor : this.runningSensors.entrySet()) {
            copy.put(sensor.getKey(), sensor.getValue());
        }
        
        if (pred) {
            startReasoningEngine(this.runningReasoner);
            for (Map.Entry<ReasonerDimension, String> entry : copy.entrySet()) {
                startSensor(entry.getKey(), entry.getValue());
            }
        } else {
            suspendReasoningEngine(this.runningReasoner);
            for (Map.Entry<ReasonerDimension, String> entry : copy.entrySet()) {
                suspendSensor(entry.getKey(), entry.getValue());
            }
        }
    }
    
    /**
     * Determines if the thread is running in silent mode.
     * 
     * @return                      <code>true</code> if the worker is running in silent mode, otherwise 
     *                              <code>false</code>.
     */
    public final boolean isSilent() {
        return !this.active;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     *
     * @return                      <code>true</code> if {@link #shutDown()} wasn't called, otherwise 
     *                              <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }

    /**
     * Determines if the thread has ended its {@link #run()} method.
     *
     * @return                      <code>true</code> if the thread has ended, otherwise <code>false</code>.
     */
    public final boolean ended() {
        return this.ended;
    }

    /**
     * General run method which calls the initialize and end methods and coordinates the different components.
     *
     * @see java.lang.Thread#run()
     */
    @Override
    public final void run() {
        init();

        while (this.isActive()) { }

        end();
        OutputUtil.info("Stopped!");
        this.ended = true;
    }

    /**
     * This method opens the required ActiveMQ connections. Furthermore it starts the property and registry service.
     */
    private void init() {
        OutputUtil.info("Thread started..");
        
        // Connect to global control topic
        globalControlTopic      = ConnectionUtil.getPublisher(MainConfig.get("LP_CONTROL_TOPIC"), this);

        // Get property service
        String descriptionPath  = MainConfig.get("REACHABILITY_AGENT_PARAMETER_DESCRIPTION_PATH");
        propertyService         = new PropertyService(this, ConfigType.SYSTEM, descriptionPath);
        propertyService.start();
        
        // Start registration service
        registryService         = new Registry(this);
        registryService.start();
    }
    
    /**
     * Starts a certain reasoner agent by sending a <i>start</i> control message.
     * 
     * @param reasoner              The reasoner agents name as {@link String}.
     */
    private void startReasoningEngine(final String reasoner) {
        if (this.runningReasoner != null && !this.runningReasoner.equals(reasoner)) {
            stopReasoningEngine(this.runningReasoner);
        }
        
        if (!this.isSilent()) {
            String id = ProjectInfo.REACHABILITY_AGENT_NAME + "_Reasoner_" + reasoner;
            sendControlMessage(ConfigOrder.start, id);
        }
        this.runningReasoner = reasoner;
        MainConfig.setProperty("REASONER", reasoner);
    }
    
    /**
     * Suspends a certain reasoner agent by sending a <i>sleep</i> control message.
     * 
     * @param reasoner              The reasoner agents name as {@link String}.
     */
    private void suspendReasoningEngine(final String reasoner) {
        String id = ProjectInfo.REACHABILITY_AGENT_NAME + "_Reasoner_" + reasoner;
        sendControlMessage(ConfigOrder.sleep, id);
    }
    
    /**
     * Stops a certain reasoner agent by sending a <i>sleep</i> control message.
     * 
     * @param reasoner              The reasoner agents name as {@link String}.
     */
    private void stopReasoningEngine(final String reasoner) {
        String id = ProjectInfo.REACHABILITY_AGENT_NAME + "_Reasoner_" + reasoner;
        sendControlMessage(ConfigOrder.sleep, id);
        this.runningReasoner = null;
    }
    
    /**
     * Starts a certain sensor agent by sending a <i>start</i> control message.
     * 
     * @param dimension             The sensor agents supported context dimension as {@link ReasonerDimension}.
     * @param sensor                The sensor agents name as {@link String}.
     */
    private void startSensor(final ReasonerDimension dimension,
                             final String sensor) {
        if (this.runningSensors.containsKey(dimension) && !this.runningSensors.get(dimension).equals(sensor)) {
            stopSensor(dimension, this.runningSensors.get(dimension));
        }
        
        if (!this.isSilent()) {
            String id = ProjectInfo.REACHABILITY_AGENT_NAME + "_Sensor_" + dimension + "_" + sensor;
            sendControlMessage(ConfigOrder.start, id);
        }
        this.runningSensors.put(dimension, sensor);
        MainConfig.setProperty("SENSOR_" + dimension, sensor);
    }
    
    /**
     * Suspends a certain sensor agent by sending a <i>sleep</i> control message.
     * 
     * @param dimension             The sensor agents supported context dimension as {@link ReasonerDimension}.
     * @param sensor                The sensor agents name as {@link String}.
     */
    private void suspendSensor(final ReasonerDimension dimension,
                               final String sensor) {
        String id = ProjectInfo.REACHABILITY_AGENT_NAME + "_Sensor_" + dimension + "_" + sensor;
        sendControlMessage(ConfigOrder.sleep, id);
    }
    
    /**
     * Stops a certain sensor agent by sending a <i>sleep</i> control message.
     * 
     * @param dimension             The sensor agents supported context dimension as {@link ReasonerDimension}.
     * @param sensor                The sensor agents name as {@link String}.
     */
    private void stopSensor(final ReasonerDimension dimension,
                            final String sensor) {
        String id = ProjectInfo.REACHABILITY_AGENT_NAME + "_Sensor_" + dimension + "_" + sensor;
        sendControlMessage(ConfigOrder.sleep, id);
        this.runningSensors.remove(dimension);
    }
    
    /**
     * Sends a new control message.
     * 
     * @param order                 The {@link ConfigOrder} for control message.
     * @param id                    The recipients ID which shall receive the control message.
     */
    private void sendControlMessage(final ConfigOrder order,
                                    final String id) {
        JsonConfigOrder msg = new JsonConfigOrder();
        
        msg.Id      = id;
        msg.order   = order;
        msg.Version = this.getVersion();
        
        OutputUtil.debug("Sending " + order + " to '" + id + "'");
        ConnectionUtil.setMessageOnTopic(globalControlTopic, msg);
    }

    /**
     * The method will be called at the end of the method {@link #run()} and will notify all connected sensor adapters 
     * and the reasoners. Furthermore it disconnects all open ActiveMQ topics and shut down all started services. 
     */
    private void end() {
        // Shut down sensor adapters
        shutDownAllSensorAdapters();

        // Shut reasoning engine down
        shutDownReasoningEngine();

        // Shut property service down
        propertyService.shutDown();
        
        // Shut down registry
        registryService.shutDown();
        
        // Disconnect own topics
        if (globalControlTopic != null) {
            globalControlTopic.disconnect();
        }
    }
    
    /**
     * Stops all active sensor agents.
     */
    private void shutDownAllSensorAdapters() {
        Map<ReasonerDimension, String> copy = new HashMap<ReasonerDimension, String>();
        for (Map.Entry<ReasonerDimension, String> sensor : this.runningSensors.entrySet()) {
            copy.put(sensor.getKey(), sensor.getValue());
        }
        
        for (Map.Entry<ReasonerDimension, String> sensor : copy.entrySet()) {
            if (sensor.getValue() != null) {
                stopSensor(sensor.getKey(), sensor.getValue());
            } else {
                this.runningSensors.remove(sensor.getKey());
            }
        }
    }
    
    /**
     * Stops the active reasoning engine.
     */
    private void shutDownReasoningEngine() {
        if (this.runningReasoner != null) {
            stopReasoningEngine(this.runningReasoner);
        }
    }
    
    /**
     * Updates the local properties by new ones. This method will be called by the {@link PropertyService} if a new
     * configuration arises on the ActiveMQ.
     * 
     * @param newProps              The new {@link Properties} set.
     * @param changedProps          Only the changed {@link Properties}.
     * @param type                  The type of properties as {@link ConfigType}.
     */
    public final void updateProperties(final Properties newProps,
                                       final Properties changedProps,
                                       final ConfigType type) {
        
        // Fetch current config
        switch(type) {
            case SYSTEM:
                // Update changed properties
                for (Map.Entry<Object, Object> entry : changedProps.entrySet()) {
                    String key      = String.valueOf(entry.getKey());
                    String marker   = "SENSOR_";
                    
                    // Changed sensor
                    if (key.startsWith(marker)) {
                        String sensor    = String.valueOf(entry.getValue());
                        String dimension = key.substring(marker.length());
                        if (EnumerationUtil.contains(ReasonerDimension.class, dimension)) {
                            ReasonerDimension theDimension = ReasonerDimension.valueOf(dimension);
                            if (this.registeredSensors.containsKey(theDimension)
                             && this.registeredSensors.get(theDimension).contains(sensor)) {
                                startSensor(theDimension, sensor);
                            }
                        }
                        
                    }
                    
                    // Changed reasoner
                    if ("REASONER".equals(key)) {
                        String reasoner = String.valueOf(entry.getValue());
                        if (this.registeredReasoners.contains(reasoner)) {
                            startReasoningEngine(reasoner);
                        }
                    }
                }
                
                // Change properties
                MainConfig.updateProperties(newProps);
                
                // Forward config to other components
                this.propertyService.pushProperties(null);
                break;
            case LOGGER:
            default:
                break;
        }
    }

    /**
     * @see PropertyServiceNotifiable#loadCurrentProperties(ConfigType)
     * 
     * @param propertiesType        The {@link ConfigType} for which the local properties shall be loaded an returned.
     * @return                      The properties if known.
     */
    public final Properties loadCurrentProperties(final ConfigType propertiesType) {
        Properties props = new Properties();
        
        switch(propertiesType) {
            case SYSTEM:
                props = MainConfig.getChangeableProperties();
                break;
            case LOGGER:
            default:
                break;
        }
        
        return props;
    }
    
    /**
     * @see PropertyServiceNotifiable#getValuesForProperty(String)
     * 
     * @param property              The name of the requested property.
     * @return                      The dynamic set of values for this property.
     */
    public final Map<String, String> getValuesForProperty(final String property) {
        Map<String, String> result = new HashMap<String, String>();

        // Reasoners
        if ("REASONER".equals(property)) {
            for (String reasoner : this.registeredReasoners) {
                result.put(reasoner, reasoner);
            }
        }
        
        // Sensors
        String marker = "SENSOR_";
        if (property.startsWith(marker)
            && EnumerationUtil.contains(ReasonerDimension.class, property.substring(marker.length()))) {
         
            ReasonerDimension dimension = ReasonerDimension.valueOf(property.substring(marker.length()));
            if (this.registeredSensors.containsKey(dimension)) {
                for (String value : this.registeredSensors.get(dimension)) {
                    result.put(value, value);
                }
            }
        }
        
        return result;
    }
    
    /**
     * Registers a new component. This may be a sensor or reasoner agent.
     * 
     * @param type                  The component type as value of {@link RegistryAgentType}.
     * @param name                  The name of the new component (the ID on which it reacts on control messages).
     * @param dimension             The supported context dimension (required for {@link RegistryAgentType#SENSOR}.
     */
    public final void register(final RegistryAgentType type,
                               final String name,
                               final ReasonerDimension dimension) {
        // Info Output
        String dim = "";
        if (dimension != null) {
            dim = " of dimension '" + dimension + "'";
        }
        OutputUtil.info("+ Registered a " + type + ": '" + name + "'" + dim);
        
        // Handle register request
        switch(type) {
            case REASONER:
                if (!this.registeredReasoners.contains(name)) {
                    this.registeredReasoners.add(name);
                }
                if (this.runningReasoner == null) {
                    startReasoningEngine(name);
                }
                break;
            case SENSOR:
                if (!this.registeredSensors.get(dimension).contains(name)) {
                    this.registeredSensors.get(dimension).add(name);
                }
                if (!this.runningSensors.containsKey(dimension)) {
                    startSensor(dimension, name);
                }
                break;
            default:
                break;
        }
    }
    
    /**
     * Unregisters a component (sensor or reasoner agent) and handles its consequences.
     * 
     * @param type                  The component type as value of {@link RegistryAgentType}.
     * @param name                  The name of the leaving component (the ID on which it reacts on control messages).
     * @param dimension             The supported context dimension (required for {@link RegistryAgentType#SENSOR}.
     */
    public final void unregister(final RegistryAgentType type,
                                 final String name,
                                 final ReasonerDimension dimension) {
        // Info Output
        String dim = "";
        if (dimension != null) {
            dim = " of dimension '" + dimension + "'";
        }
        OutputUtil.info("- Unregistered a " + type + ": '" + name + "'" + dim);
        
        // Handle unregister request
        switch(type) {
            case REASONER:
                this.runningReasoner = null;
                
                if (this.registeredReasoners.contains(name)) {
                    this.registeredReasoners.remove(name);
                }
                
                if (!this.registeredReasoners.isEmpty()) {
                    startReasoningEngine((String) this.registeredReasoners.toArray()[0]);
                }
                break;
            case SENSOR:
                this.runningSensors.remove(dimension);
                
                if (this.registeredSensors.get(dimension).contains(name)) {
                    this.registeredSensors.get(dimension).remove(name);
                }
                
                if (!this.registeredSensors.get(dimension).isEmpty()) {
                    startSensor(dimension, (String) this.registeredSensors.get(dimension).toArray()[0]);
                }
                break;
            default:
                break;
        }
    }

    /**
     * Returns the agents version.
     *
     * @return                      The agents version as {@link ProjectInfo#REACHABILITY_AGENT_VERSION}.
     */
    private String getVersion() {
        return ProjectInfo.REACHABILITY_AGENT_VERSION;
    }
}
