package de.hawhamburg.reachability.util;

import java.io.IOException;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.Level;

import de.hawhamburg.lifecycle.data.ConfigOrder;
import de.hawhamburg.lifecycle.data.JsonConfigOrder;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.LoggerConfig;
import de.hawhamburg.reachability.conf.MainConfig;

/**
 * Utility functions related to information output.<br><br>
 * 
 * This utility class provides some fancy utility methods related to output matters (e.g. console or logfile).
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public final class OutputUtil {
    
    /**
     * A {@link Logger} object for logfile output. This object will be only set once a programs life time.
     */
    private static Logger logger    = null;
    
    /**
     * An information flag which indicates if it will be possible to log messages to a file using the {@link #logger}. 
     * It will be set to <code>false</code> if the method {@link #configLogger()} is, for some reasons, unable to create
     *  the specified logfile.
     */
    private static boolean logable  = true;

    /**
     * Prevent instances of this utility class.
     */
    private OutputUtil() { }
    
    /**
     * Handles a debug messages depending on the current configuration in {@link MainConfig}. This method will only work
     * if <i>DEBUG</i> is set to <code>true</code>.<br><br>
     * 
     * If <i>PRINT_OUTPUT</i> is set to <code>true</code>, the message will printed to console (prefixed by the class 
     * name, if it isn't <code>null</code>).<br><br>
     * 
     * If <i>LOG_OUTPUT</i> is set to <code>true</code>, the message will be logged to the logfile with level 
     * <i>WARNING</i> and the class name, if it isn't <code>null</code>.
     * 
     * @param msg           A debug message as {@link String}.
     */
    public static void debug(final String msg) {
        if (MainConfig.get("DEBUG", Boolean.class)) {
            handleMessage(msg, Level.WARNING);
        }
    }
    
    /**
     * Handles an info messages depending on the current configuration in {@link MainConfig}. This method will only work
     * if <i>INFO</i> is set to <code>true</code>.<br><br>
     * 
     * If <i>PRINT_OUTPUT</i> is set to <code>true</code>, the message will printed to console (prefixed by the class 
     * name, if it isn't <code>null</code>).<br><br>
     * 
     * If <i>LOG_OUTPUT</i> is set to <code>true</code>, the message will be logged to the logfile with level 
     * <i>INFO</i> and the class name, if it isn't <code>null</code>.
     * 
     * @param msg           An info message as {@link String}.
     */
    public static void info(final String msg) {
        if (MainConfig.get("INFO", Boolean.class)) {
            handleMessage(msg, Level.INFO);
        }
    }
    
    /**
     * Handles an error messages depending on the current configuration in {@link MainConfig}. This method will only 
     * work if <i>ERROR</i> is set to <code>true</code>. <br><br>
     * 
     * If <i>PRINT_OUTPUT</i> is set to <code>true</code>, the message will printed as error to console (prefixed by the
     * class name, if it isn't <code>null</code>).<br><br>
     * 
     * If <i>LOG_OUTPUT</i> is set to <code>true</code>, the message will be logged to the logfile with level 
     * <i>SEVERE</i> and the class name, if it isn't <code>null</code>.
     * 
     * @param msg           An error message as {@link String}. 
     */
    public static void error(final String msg) {
        if (MainConfig.get("ERROR", Boolean.class)) {
            handleError(msg, Level.SEVERE);
        }
    }
    
    /**
     * Handles the error message using {@link #handleError(String, Level)}. Sends a <b>stop</b> order on the 
     * <i>LP_CONTROL_TOPIC</i> using the JSON message {@link JsonConfigOrder} to itself to gracefully end this agent.
     * 
     * @param msg           An error message as {@link String}.
     */
    public static void endSystemWithError(final String msg) {
        handleError(msg, Level.SEVERE);
        
        // Set up stopping order to end this agent
        JsonConfigOrder order = new JsonConfigOrder();
        order.Version   = ProjectInfo.REACHABILITY_AGENT_VERSION;
        order.Id        = ProjectInfo.REACHABILITY_AGENT_NAME;
        order.order     = ConfigOrder.stop;
        
        // Send stopping order
        LPPublisher ctrl = ConnectionUtil.getPublisher(MainConfig.get("LP_CONTROL_TOPIC"), null);
        ConnectionUtil.setMessageOnTopic(ctrl, order);
        ctrl.disconnect();
    }

    /**
     * Handles an info or debug message depending on the current configuration in {@link MainConfig}.<br><br>
     * 
     * If <i>PRINT_OUTPUT</i> is set to <code>true</code>, the message will printed to console (prefixed by the class 
     * name or <i>Unknown</i> if it's <code>null</code>).<br><br>
     * 
     * If <i>LOG_OUTPUT</i> is set to <code>true</code>, the message will be logged to the logfile with the specified 
     * {@link Level} and the class name or <i>Unknown</i> if it's <code>null</code>.
     * 
     * @param msg           An info or debug message as {@link String}.
     * @param lvl           A logging {@link Level}.
     */
    public static void handleMessage(final String msg,
                                     final Level lvl) {
        String msgOut = buildMessage(msg);
        
        if (MainConfig.get("PRINT_OUTPUT", Boolean.class)) {
            System.out.println(msgOut);
        }
        if (MainConfig.get("LOG_OUTPUT", Boolean.class)) {
            log(lvl, msgOut);
        }
    }
    
    /**
     * Handles an error message depending on the current configuration in {@link MainConfig}.<br><br>
     * 
     * If <i>PRINT_OUTPUT</i> is set to <code>true</code>, the message will printed as error to console (prefixed by the
     * class name or <i>Unknown</i> if it's <code>null</code>).<br><br>
     * 
     * If <i>LOG_OUTPUT</i> is set to <code>true</code>, the message will be logged to the logfile with the specified 
     * {@link Level} and the class name or <i>Unknown</i> if it's 
     * <code>null</code>.
     * 
     * @param msg           An error message as {@link String}.
     * @param lvl           A logging {@link Level}.
     */
    public static void handleError(final String msg,
                                   final Level lvl) {
        String msgOut = buildMessage(msg);
        
        if (MainConfig.get("PRINT_OUTPUT", Boolean.class)) {
            System.err.println(msgOut);
        }
        if (MainConfig.get("LOG_OUTPUT", Boolean.class)) {
            log(lvl, msgOut);
        }
    }
    
    /**
     * Builds a message from the given message. This method determines the original method and class from which this 
     * utility class has been called and integrates them into the result message.
     *   
     * @param msg           A message as {@link String}.
     * @return              A {@link String} containing the originators class and method name.
     */
    private static String buildMessage(final String msg) {
        
        // Try to select the method name:
        StackTraceElement[] trace = Thread.currentThread().getStackTrace();
        StackTraceElement source  = null;
        for (int i = 1; i < trace.length; i++) {
            if (!trace[i].getClassName().equals(OutputUtil.class.getName())) {
                source = trace[i];
                break;
            }
        }
        
        return source + "| " + msg;
    }
    
    /**
     * Logs the specified message with a certain loggig {@link Level} to the logfile by using the {@link #logger}, if 
     * {@link #logable} is set to <code>true</code>. If no logger is configured, the method will call 
     * {@link #configLogger()} to create on.   
     * 
     * @param lvl       A logging {@link Level}.
     * @param msg       A message as {@link String} which shall be logged to the logfile.
     */
    public static void log(final Level lvl,
                           final String msg) {
        if (logable) {
            if (logger == null) {
                configLogger();
            }
            if (logger != null) {
                logger.log(lvl, msg);
            }
        }
    }
    
    /**
     * Sets up a logger without console output with name <i>LOGGER_NAME</i>. The logger will write its logging to the 
     * logfile <i>LOGGER_FILE</i> which will grow up to <i>LOGGER_MAX_FILE_SIZE</i>s. The logging level 
     * {@link Level#WARNING} will be written as <i>DEBUG</i>. If it is not possible to create the logfile, the whole 
     * system will end itself controlled with console output.
     */
    private static void configLogger() {
        // Remove console output from default logger
        Logger root = Logger.getLogger("");
        Handler[] handlers = root.getHandlers();
        if (handlers.length > 0 && handlers[0] instanceof ConsoleHandler) {
            root.removeHandler(handlers[0]);
        }
        
        // Create file handler
        if (LoggerConfig.exists()) {
            try {
                // Get configuration
                final String name         = LoggerConfig.get("LOGGER_NAME");
                final String file         = LoggerConfig.get("LOGGER_FILE");
                final int    maxFileSize  = LoggerConfig.get("LOGGER_MAX_FILE_SIZE", Integer.class);
                final int    maxFiles     = LoggerConfig.get("LOGGER_MAX_FILES", Integer.class);
                
                // Set-up a file handler for logging
                FileHandler logfile = new FileHandler(
                        file.replaceAll("__TIME__", "" + System.currentTimeMillis()),
                        maxFileSize, maxFiles);
                
                // Re-write file handler formatter for customized output
                logfile.setFormatter(new Formatter() {
                    public String format(final LogRecord rec) {
                        StringBuffer buf = new StringBuffer(maxFileSize);
                        String lvl;
                        if (rec.getLevel() == Level.WARNING) {
                            lvl = "DEBUG";
                        } else {
                            lvl = rec.getLevel().toString();
                        }
                        buf.append(new java.util.Date()).
                            append(' ').
                            append(lvl).
                            append(' ').
                            append(formatMessage(rec)).
                            append('\n');
                        return buf.toString();
                    }
                });
                logger = Logger.getLogger(name);
                logger.addHandler(logfile);
            } catch (SecurityException e) {
                logable = false;
                endSystemWithError("Unable to create a logging file: " + e.getMessage());
            } catch (IOException e) {
                logable = false;
                endSystemWithError("Unable to create a logging file: " + e.getMessage());
            }
        } else {
            MainConfig.setProperty("LOG_OUTPUT", false);
            OutputUtil.error("No logger configurations found. Logger deactivated!");
        }
    }
}
